﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  06/26/13    Created 
//
// The SGI Image File Format
// Specification ftp://ftp.sgi.com/graphics/grafica/sgiimage.html
//---------------------------------------------------------------------------

namespace System.Drawing.Imaging
{
    using System.IO;

    /// <summary>
    /// Decodes bitmaps using the SGI image file format.
    /// </summary>
    public sealed class SGIDecoder : SGICodec, IBitmapDecoder
    {
        #region Constants
        /// <summary>
        /// Defines a string indicating the SGI dimension is not supported.
        /// </summary>
        private const string UNSUPPORTED_DIMENSION_EXCEPTION = "SGI DIMENSION '{0}' is not supported.";
        /// <summary>
        /// Defines a string indicating the number of SGI channels is not supported.
        /// </summary>
        private const string UNSUPPORTED_ZSIZE_EXCEPTION = "SGI ZSIZE '{0}' is not supported.";
        /// <summary>
        /// Defines a string indicating the SGI storage (compression) is not supported.
        /// </summary>
        private const string UNSUPPORTED_STORAGE_EXCEPTION = "SGI STORAGE '{0}' is not supported.";
        /// <summary>
        /// Defines a string indicating the BCP (number of bytes per pixel channel) has a value not implemented.
        /// </summary>
        private const string NONIMPLEMENTED_BCP_EXCEPTION = "SGI BCP '{0}' is not implemented.";
        /// <summary>
        /// Defines a string indicating the BCP (number of bytes per pixel channel) is not supported.
        /// </summary>
        private const string UNSUPPORTED_BCP_EXCEPTION = "SGI BCP '{0}' is not supported.";
        /// <summary>
        /// Defines a string indicating the SGI color map is not supported.
        /// </summary>
        private const string UNSUPPORTED_COLORMAP_EXCEPTION = "SGI COLORMAP '{0}' is not supported.";
        /// <summary>
        /// Defines a string indicating the SGI color map is not implemented.
        /// </summary>
        private const string NONIMPLEMENTED_COLORMAP_EXCEPTION = "SGI COLORMAP '{0}' is not implemented.";
        /// <summary>
        /// Defines a string indicating the SGI minimum pixel is not implemented.
        /// </summary>
        private const string UNSUPPORTED_PIXMIN_EXCEPTION = "SGI PIXMIN '{0}' is not implemented.";
        /// <summary>
        /// Defines a string indicating the SGI maximum pixel is not implemented.
        /// </summary>
        private const string UNSUPPORTED_PIXMAX_EXCEPTION = "SGI PIXMAX '{0}' is not implemented.";
        #endregion

        #region Methods
        /// <summary>
        /// Decodes an SGI bitmap from the specified reader.
        /// </summary>
        /// <param name="reader">The reader to read binary bitmap data from.</param>
        /// <returns>A Bitmap read from the underlying stream.</returns>
        public Bitmap Decode(BinaryReader reader)
        {
            uint start = (uint)reader.BaseStream.Position; //store position to allow agnostic bitmap positions

            short MAGIC = reader.ReadBigEndianInt16(); //read magic number

            if (MAGIC != MAGIC_NUMBER) //validate magic number
                throw new InvalidFileType(MAGIC_NUMBER.ToString(), MAGIC.ToString());

            SGIEncoding STORAGE = (SGIEncoding)reader.ReadByte();

            byte BPC = reader.ReadByte(); //BCP pixel precision

            //validate pixel precision:
            switch (BPC)
            {
                case 1:
                    break;
                case 2:
                    throw new NotImplementedException(string.Format(NONIMPLEMENTED_BCP_EXCEPTION, BPC));
                default:
                    throw new InvalidFileDataException(string.Format(UNSUPPORTED_BCP_EXCEPTION, BPC));
            }

            ushort DIMENSION = reader.ReadBigEndianUInt16(); //Number of dimensions
            ushort XSIZE = reader.ReadBigEndianUInt16(); //width
            ushort YSIZE = reader.ReadBigEndianUInt16(); //height
            ushort ZSIZE = reader.ReadBigEndianUInt16(); //channels
            uint PIXMIN = reader.ReadBigEndianUInt32(); //Minimum pixel value 

            if (PIXMIN != 0)
                throw new NotImplementedException(string.Format(UNSUPPORTED_PIXMIN_EXCEPTION, PIXMIN));

            uint PIXMAX = reader.ReadBigEndianUInt32(); //Maximum pixel value

            if (PIXMAX != 255)
                throw new NotImplementedException(string.Format(UNSUPPORTED_PIXMAX_EXCEPTION, PIXMAX));

            int DUMMY = reader.ReadBigEndianInt32(); //Ignored
            byte[] IMAGENAME = reader.ReadBytes(80); //Image name
            uint COLORMAP = reader.ReadBigEndianUInt32(); //Colormap ID

            //validate COLORMAP:
            switch (COLORMAP)
            {
                case 0: //only implemented NORMAL color maps currently
                    break;
                case 1: //DITHERED 
                case 2: //SCREEN 
                case 3: //COLORMAP
                    throw new NotImplementedException(string.Format(NONIMPLEMENTED_COLORMAP_EXCEPTION, COLORMAP));
                default:
                    throw new InvalidFileDataException(string.Format(UNSUPPORTED_COLORMAP_EXCEPTION, COLORMAP));
            }

            byte[] DUMMYB = reader.ReadBytes(404); //Ignored

            SGIFormat format;

            //validates and determines the format:
            switch (DIMENSION)
            {
                case 1: //single channel and single scanline
                case 2: //single channel and multiple scanlines
                    format = SGIFormat.BW;
                    break;
                case 3: //multiple channels and multiple scanlines
                    switch (ZSIZE)
                    {
                        case 2: //BW+A
                            format = SGIFormat.BWA;
                            break;
                        case 3: //RGB
                        case 4: //RGBA
                            format = (ZSIZE == 4 ? SGIFormat.RGBA : SGIFormat.RGB);
                            break;
                        default:
                            throw new NotSupportedException(string.Format(UNSUPPORTED_ZSIZE_EXCEPTION, ZSIZE));
                    }
                    break;
                default:
                    throw new NotSupportedException(string.Format(UNSUPPORTED_DIMENSION_EXCEPTION, DIMENSION));
            }

            Bitmap bitmap = null;

            try
            {
                //initialize the bitmap:
                switch (format)
                {
                    case SGIFormat.BWA:
                    case SGIFormat.RGBA:
                        bitmap = new Bitmap(XSIZE, YSIZE, PixelFormat.Format32bppArgb);
                        break;
                    default:
                        bitmap = new Bitmap(XSIZE, YSIZE, PixelFormat.Format32bppRgb);
                        break;
                }

                //initialize the BitmapPixels class to do fast pixel processing
                using (BitmapPixels pixels = new BitmapPixels(bitmap))
                {
                    //lock the bitmap:
                    pixels.Lock();

                    unsafe
                    {
                        switch (STORAGE)
                        {
                            case SGIEncoding.RLE: //decodes run length encoded bitmap

                                int scanlines = YSIZE * ZSIZE; //number of expected scanlines

                                //scanline stream start positions:
                                uint[] starts = new uint[scanlines];

                                //read and offset each start position by the start position:
                                for (int i = 0; i < scanlines; i++)
                                    starts[i] = reader.ReadBigEndianUInt32() + start;

                                //read the scanline length (unused currently), just skip bytes:
                                //for (int i = 0; i < scanlines; i++)
                                //reader.ReadBigEndianUInt32();
                                reader.BaseStream.Seek(scanlines * 4, SeekOrigin.Current);

                                for (int y = 0; y < YSIZE; y++) //read each row
                                {
                                    //SGI images are stored from bottom to top while Bitmap pixels are stored top to bottom.
                                    //this offsets the pointer to the correct row:
                                    byte* buffer_p = (byte*)(pixels.Pointer + ((YSIZE - 1) - y) * XSIZE);

                                    for (int z = 0; z < ZSIZE; z++)
                                    {
                                        //offset the pointer from BGRA to RGBA:
                                        byte* buffer_ps = OffsetChannelPointer(format, z, buffer_p);

                                        //reads run length encoded channel scanline:
                                        ReadRLEScanline(reader.BaseStream, starts[y + YSIZE * z], buffer_ps);
                                    }

                                    //BW images need to have the color component copied:
                                    switch (format)
                                    {
                                        case SGIFormat.BW:
                                        case SGIFormat.BWA:

                                            RGBAColor* buffer_ps = (RGBAColor*)buffer_p;

                                            for (int x = 0; x < XSIZE; x++)
                                                CopyColorComponents(ref buffer_ps);

                                            break;
                                    }
                                }
                                break;
                            case SGIEncoding.Verbatim: //reads bitmap pixels

                                for (int z = 0; z < ZSIZE; z++)
                                {
                                    for (int y = 0; y < YSIZE; y++)
                                    {
                                        //SGI images are stored from bottom to top while Bitmap pixels are stored top to bottom.
                                        //this offsets the pointer to the correct row:
                                        byte* buffer_p = (byte*)(pixels.Pointer + ((YSIZE - 1) - y) * XSIZE);

                                        //offset the pointer from BGRA to RGBA:
                                        byte* buffer_ps = OffsetChannelPointer(format, z, buffer_p);

                                        //reads channel scanline verbatim:
                                        for (int x = 0; x < XSIZE; x++)
                                        {
                                            *buffer_ps = reader.ReadByte();
                                            buffer_ps += 4;
                                        }
                                    }
                                }

                                //BW images need to have the color component copied:
                                switch (format)
                                {
                                    case SGIFormat.BW:
                                    case SGIFormat.BWA:

                                        RGBAColor* buffer_ps = pixels.Pointer;

                                        for (int i = 0; i < YSIZE * XSIZE; i++)
                                            CopyColorComponents(ref buffer_ps);

                                        break;
                                }

                                break;
                            default:
                                throw new NotSupportedException(string.Format(UNSUPPORTED_STORAGE_EXCEPTION, STORAGE));
                        }
                    }
                }

                return bitmap;
            }
            catch
            {
                if (bitmap != null)
                    bitmap.Dispose();

                throw;
            }
        }

        /// <summary>
        /// Copies the blue color component to each channel except alpha.
        /// </summary>
        /// <param name="buffer_ps">A pointer to bitmap pixels.</param>
        private static unsafe void CopyColorComponents(ref RGBAColor* buffer_ps)
        {
            *buffer_ps = new RGBAColor(buffer_ps->A << 24 | buffer_ps->B << 16 | buffer_ps->B << 8 | buffer_ps->B);
            buffer_ps++;
        }

        /// <summary>
        /// Reads a scanline from the stream into a pointer.
        /// </summary>
        /// <param name="stream">The stream to read from.</param>
        /// <param name="start">The stream start position.</param>
        /// <param name="buffer">A pointer to copy data to.</param>
        private static unsafe void ReadRLEScanline(Stream stream, uint start, byte* buffer)
        {
            //set the stream position:
            stream.Position = start;

            int x = 0;
            int pixel;
            int count;

            for (; ; )
            {
                pixel = stream.ReadByte(); //read the first value
                count = pixel & 127; //gets the number of pixels to read or copy

                if (count == 0)
                    break;

                if ((pixel & 0x80) > 0) //determines if pixels should be read
                {
                    while (count > 0) //read count pixels
                    {
                        *buffer = (byte)stream.ReadByte(); //set pixel to buffer

                        count--; //decrement count
                        x++; //increment horizontal position
                        buffer += 4; //get next pixel
                    }
                }
                else //otherwise, pixels should be copied
                {
                    pixel = stream.ReadByte(); //read first pixel

                    while (count > 0) //copy count times
                    {
                        *buffer = (byte)pixel; //set pixel

                        count--; //decrement count
                        x++; //increment horizontal position
                        buffer += 4; //get next pixel
                    }
                }
            }
        }
        #endregion
    }
}