﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  06/26/13    Created 
//
// The SGI Image File Format
// Specification ftp://ftp.sgi.com/graphics/grafica/sgiimage.html
//---------------------------------------------------------------------------

namespace System.Drawing.Imaging
{
    using System.IO;

    /// <summary>
    /// Encodes bitmaps using the SGI image file format.
    /// </summary>
    public sealed class SGIEncoder : SGICodec, IBitmapEncoder
    {
        #region Constants
        /// <summary>
        /// Defines a string indicating an invalid SGIFormat was specified.
        /// </summary>
        private const string INVALID_FORMAT_EXCEPTION = "The specified format: '{0}' is not supported.";
        /// <summary>
        /// Defines a string indicating an invalid SGIEncoding was specified.
        /// </summary>
        private const string INVALID_ENCODING_EXCEPTION = "The specified encoding: '{0}' is not supported.";
        #endregion

        #region Fields
        /// <summary>
        /// Defines the SGI color format.
        /// </summary>
        public readonly SGIFormat Format;
        /// <summary>
        /// Defines the SGI binary encoding.
        /// </summary>
        public readonly SGIEncoding Encoding;
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes the encoder with a format and encoding.
        /// </summary>
        /// <param name="format">The SGI color format to encode.</param>
        /// <param name="encoding">The SGI binary encoding.</param>
        public SGIEncoder(SGIFormat format, SGIEncoding encoding)
        {
            switch (format)
            {
                case SGIFormat.BW:
                case SGIFormat.BWA:
                case SGIFormat.RGB:
                case SGIFormat.RGBA:
                    break;
                default:
                    throw new ArgumentException(string.Format(INVALID_FORMAT_EXCEPTION, format), "format");
            }

            switch (encoding)
            {
                case SGIEncoding.RLE:
                case SGIEncoding.Verbatim:
                    break;
                default:
                    throw new ArgumentException(string.Format(INVALID_ENCODING_EXCEPTION, encoding), "encoding");
            }

            Format = format;
            Encoding = encoding;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Encodes the bitmap and writes to the specified writer.
        /// </summary>
        /// <param name="bitmap">The bitmap to encode and write to the underlying stream.</param>
        /// <param name="writer">A writer to write to.</param>
        public void Encode(Bitmap bitmap, BinaryWriter writer)
        {
            if (bitmap == null)
                throw new ArgumentNullException("bitmap");

            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteBigEndian(MAGIC_NUMBER); //write magic number

            switch (Encoding)
            {
                case SGIEncoding.Verbatim:
                case SGIEncoding.RLE:
                    writer.Write((byte)Encoding); //write encoding
                    break;
                default:
                    throw new NotSupportedException(string.Format(INVALID_ENCODING_EXCEPTION, Encoding));
            }

            writer.Write((byte)1); //bpc

            ushort XSIZE = (ushort)bitmap.Width;
            ushort YSIZE = (ushort)bitmap.Height;

            ushort ZSIZE;
            ushort DIMENSION;

            //determine number of channels and dimensions:
            switch (Format)
            {
                case SGIFormat.RGBA:
                    ZSIZE = 4;
                    DIMENSION = 3;
                    break;
                case SGIFormat.RGB:
                    ZSIZE = 3;
                    DIMENSION = 3;
                    break;
                case SGIFormat.BWA:
                    ZSIZE = 2;
                    DIMENSION = 3;
                    break;
                case SGIFormat.BW:
                    ZSIZE = 1;
                    DIMENSION = YSIZE == 1 ? (ushort)1 : (ushort)2;
                    break;
                default:
                    throw new NotSupportedException(string.Format(INVALID_FORMAT_EXCEPTION, Format));
            }

            writer.WriteBigEndian(DIMENSION); //dimension
            writer.WriteBigEndian(XSIZE); //width
            writer.WriteBigEndian(YSIZE); //height
            writer.WriteBigEndian(ZSIZE); //number of channels

            writer.WriteBigEndian((uint)0); //PIXMIN
            writer.WriteBigEndian((uint)255); //PIXMAX
            writer.Write(new byte[4]); //DUMMY
            writer.Write(new byte[80]); //IMAGENAME
            writer.WriteBigEndian((uint)0); //COLORMAP : normal mode
            writer.Write(new byte[404]); //DUMMY

            uint[] starts;
            uint[] lengths;
            long scanlineStreamPosition;

            if (Encoding == SGIEncoding.RLE)
            {
                int scanlines = YSIZE * ZSIZE;

                //initialize start and length arrays to hold scanline stream positions and lengths:
                starts = new uint[scanlines];
                lengths = new uint[scanlines];

                scanlineStreamPosition = writer.BaseStream.Position;

                //write place holders for each row in each channel:
                writer.Write(new byte[4 * scanlines * 2]); //4 bytes * number of scanlines for start and lengths
            }
            else
            {
                //initialize unused locals:
                starts = null;
                lengths = null;
                scanlineStreamPosition = 0;
            }

            using (BitmapPixels pixels = new BitmapPixels(bitmap))
            {
                pixels.Lock();

                unsafe
                {
                    for (int z = 0; z < ZSIZE; z++) //writes each channel
                    {
                        for (int y = 0; y < YSIZE; y++) //writes each channel scanline
                        {
                            //SGI images are stored from bottom to top while Bitmap pixels are stored top to bottom.
                            //this offsets the pointer to the correct row:
                            byte* buffer_p = (byte*)(pixels.Pointer + ((YSIZE - 1) - y) * XSIZE);
                            //offset the pointer from BGRA to RGBA:
                            byte* buffer_ps = OffsetChannelPointer(Format, z, buffer_p);

                            if (Encoding == SGIEncoding.RLE)
                                //store the stream position the scanline is wrote to:
                                starts[y + YSIZE * z] = (uint)writer.BaseStream.Position;

                            byte[] channel = new byte[XSIZE]; //create a temporary channel buffer

                            fixed (byte* channel_p = channel) //get a pointer to the channel buffer
                            {
                                byte* channel_ps = channel_p; //duplicate pointer

                                for (int x = 0; x < XSIZE; x++) //fills the buffer with bitmap color
                                {
                                    switch (Format) //different formats need to be handled
                                    {
                                        case SGIFormat.BW: //convert the image to BW and apply opacity
                                            *channel_ps = (byte)((*(buffer_ps + 2) * 0.2126 + *(buffer_ps + 1) * 0.7152 + *buffer_ps * 0.0722) * *(buffer_ps + 3) / 255D);
                                            break;
                                        case SGIFormat.BWA: //convert the image to BW

                                            if (z == 0) //if RGB:
                                                *channel_ps = (byte)(*(buffer_ps + 2) * 0.2126 + *(buffer_ps + 1) * 0.7152 + *buffer_ps * 0.0722);
                                            else //if A:
                                                *channel_ps = *buffer_ps;

                                            break;
                                        case SGIFormat.RGB: //apply opacity
                                            *channel_ps = (byte)(*buffer_ps * *(buffer_ps + z + 1) / 255D);
                                            break;
                                        case SGIFormat.RGBA:
                                            *channel_ps = *buffer_ps;
                                            break;
                                    }

                                    buffer_ps += 4; //gets next channel color
                                    channel_ps++; //gets next position in temporary channel buffer
                                }

                                switch (Encoding)
                                {
                                    case SGIEncoding.RLE:

                                        //run length encode the scanline:
                                        uint length;
                                        WriteRLEScanline(writer, XSIZE, channel_p, out length);
                                        lengths[y + YSIZE * z] = length; //store the length of the scanline

                                        break;
                                    case SGIEncoding.Verbatim:

                                        channel_ps = channel_p;

                                        //write out the scanline:
                                        for (int x = 0; x < XSIZE; x++)
                                        {
                                            writer.Write(*channel_ps);
                                            channel_ps++;
                                        }

                                        break;
                                }
                            }

                            if (Encoding == SGIEncoding.RLE)
                            {
                                long position = writer.BaseStream.Position; //store the current stream position

                                writer.BaseStream.Position = scanlineStreamPosition; //move to end of header

                                //write scanline start positions:
                                foreach (uint start in starts)
                                    writer.WriteBigEndian(start);

                                //write scanline data lengths:
                                foreach (uint length in lengths)
                                    writer.WriteBigEndian(length);

                                //restore stream position
                                writer.BaseStream.Position = position; //return to stored position
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Run length encodes and writes a scanline to a stream.
        /// </summary>
        /// <param name="writer">The writer to write to.</param>
        /// <param name="XSIZE">The scanline width.</param>
        /// <param name="buffer_p">A pointer to the scanline pixels.</param>
        /// <param name="length">When this method returns, contains the number of bytes wrote to the stream.</param>
        private static unsafe void WriteRLEScanline(BinaryWriter writer, int XSIZE, byte* buffer_p, out uint length)
        {
            length = 0;

            int x = 0;
            byte pixel = 0;
            int count = 0;
            int searchX = 0;
            int missed = 0;
            int writeX = 0;

            for (; ; )
            {
                if (x == XSIZE)
                    break;

                pixel = *(buffer_p + x); //get pixel
                count = 1; //set number of pixels

                //search for any identical pixels:
                for (searchX = x + 1; searchX - x < 127 && searchX < XSIZE; searchX++)
                {
                    if (*(buffer_p + searchX) == pixel)
                        count += 1;
                    else
                        break;
                }

                missed = 1; //init missed to 1 indicating a single invalid pixel

                //if there are less than 3 identical pixels, might as well check for more to write without encoding:
                if (count < 3)
                {
                    //searches for additional pixels that are not identical:
                    for (searchX = x + 1; searchX - x < 127 && searchX + 1 < XSIZE; searchX += 2)
                    {
                        if (*(buffer_p + searchX) != *(buffer_p + searchX + 1))
                            missed++;
                        else
                            break;
                    }
                }

                if (missed > 1) //if we have different pixels
                {
                    writer.Write((byte)(128 + missed)); //write flag and missed count
                    length++; //increment byte length

                    //write out different pixels:
                    for (writeX = x; writeX < x + missed; writeX++)
                    {
                        writer.Write(*(buffer_p + writeX));
                        length++; //increment byte length
                    }

                    x += missed; //increment x cursor
                }
                else
                {
                    writer.Write((byte)count); //write number of identical pixels
                    length++; //increment byte length
                    writer.Write(pixel); //write pixel
                    length++; //increment byte length
                    x += count; //increment x cursor by the number of identical pixels
                }
            }

            writer.Write((byte)0); //output terminating flag
            length++; //increment byte length
        }
        #endregion
    }
}