﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  06/26/13    Created 
//
//---------------------------------------------------------------------------

namespace System.IO
{
    using System.Drawing;
    using System.Drawing.Imaging;

    public static class BinaryReaderExtensions
    {
        /// <summary>
        /// Reads an unsigned 32-bit integer from big-endian encoded bytes.
        /// </summary>
        /// <param name="reader">The reader to read data from.</param>
        /// <returns>A 32-bit unsigned integer read from a stream.</returns>
        public static uint ReadBigEndianUInt32(this BinaryReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            return BitConverter.ToUInt32(reader.ReadBigEndianBytes(4), 0);
        }

        /// <summary>
        /// Reads a 32-bit integer from big-endian encoded bytes.
        /// </summary>
        /// <param name="reader">The reader to read data from.</param>
        /// <returns>A 32-bit signed integer read from a stream.</returns>
        public static int ReadBigEndianInt32(this BinaryReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            return BitConverter.ToInt32(reader.ReadBigEndianBytes(4), 0);
        }

        /// <summary>
        /// Reads a 16-bit integer from a stream encoded in big-endian bytes.
        /// </summary>
        /// <param name="reader">The reader to read data from.</param>
        /// <returns>A 16-bit signed integer read from a stream.</returns>
        public static short ReadBigEndianInt16(this BinaryReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            return BitConverter.ToInt16(reader.ReadBigEndianBytes(2), 0);
        }

        /// <summary>
        /// Reads an unsigned 16-bit integer from a stream encoded in big-endian bytes.
        /// </summary>
        /// <param name="reader">The reader to read data from.</param>
        /// <returns>An unsigned 16-bit signed integer read from a stream.</returns>
        public static ushort ReadBigEndianUInt16(this BinaryReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            return BitConverter.ToUInt16(reader.ReadBigEndianBytes(2), 0);
        }

        /// <summary>
        /// Reads the specified number of bytes from the reader, reversing the array if little-endian.
        /// </summary>
        /// <param name="reader">A reader to read bytes from.</param>
        /// <param name="count">The number of bytes to read.</param>
        /// <returns>A byte array containing data read from the underlying stream.</returns>
        private static byte[] ReadBigEndianBytes(this BinaryReader reader, int count)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            byte[] data = reader.ReadBytes(count);

            if (BitConverter.IsLittleEndian)
                Array.Reverse(data);

            return data;
        }

        /// <summary>
        /// Reads and decodes a bitmap.
        /// </summary>
        /// <param name="reader">A reader to read from.</param>
        /// <param name="decoder">A bitmap decoder to decode bytes.</param>
        /// <returns>A bitmap read from the underlying stream.</returns>
        public static Bitmap ReadBitmap(this BinaryReader reader, IBitmapDecoder decoder)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            if (decoder == null)
                throw new ArgumentNullException("decoder");

            return decoder.Decode(reader);
        }
    }
}
