﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  06/26/13    Created 
//
//---------------------------------------------------------------------------

namespace System.IO
{
    using System.Drawing;
    using System.Drawing.Imaging;

    public static class BinaryWriterExtensions
    {
        /// <summary>
        /// Writes a 32-bit unsigned integer in big-endian encoded bytes.
        /// </summary>
        /// <param name="writer">The stream to write to.</param>
        /// <param name="value">An 32-bit unsigned integer to write to a stream.</param>
        public static void WriteBigEndian(this BinaryWriter writer, uint value)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteBigEndianBytes(BitConverter.GetBytes(value));
        }

        /// <summary>
        /// Writes a 32-bit integer in big-endian encoded bytes.
        /// </summary>
        /// <param name="writer">The stream to write to.</param>
        /// <param name="value">An 32-bit integer to write to a stream.</param>
        public static void WriteBigEndian(this BinaryWriter writer, int value)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteBigEndianBytes(BitConverter.GetBytes(value));
        }

        /// <summary>
        /// Writes a 16-bit integer to a stream in big-endian bytes.
        /// </summary>
        /// <param name="writer">The stream to write to.</param>
        /// <param name="value">An 16-bit integer to write to a stream.</param>
        public static void WriteBigEndian(this BinaryWriter writer, short value)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteBigEndianBytes(BitConverter.GetBytes(value));
        }

        /// <summary>
        /// Writes an unsigned 16-bit integer to a stream in big-endian bytes.
        /// </summary>
        /// <param name="writer">The stream to write to.</param>
        /// <param name="value">An unsigned 16-bit integer to write to a stream.</param>
        public static void WriteBigEndian(this BinaryWriter writer, ushort value)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteBigEndianBytes(BitConverter.GetBytes(value));
        }

        /// <summary>
        /// Writes bytes to the writer, reversing the array if little-endian.
        /// </summary>
        /// <param name="writer">A writer to write bytes to.</param>
        /// <param name="data">A byte array containing the data to write.</param>
        private static void WriteBigEndianBytes(this BinaryWriter writer, byte[] data)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            if (data == null)
                throw new ArgumentNullException("data");

            if (BitConverter.IsLittleEndian)
                Array.Reverse(data);

            writer.Write(data);
        }

        /// <summary>
        /// Encodes and writes a bitmap.
        /// </summary>
        /// <param name="writer">The writer to write to.</param>
        /// <param name="bitmap">A bitmap to write.</param>
        /// <param name="encoder">The encoder to encode the bitmap with.</param>
        public static void Write(this BinaryWriter writer, Bitmap bitmap, IBitmapEncoder encoder)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            if (bitmap == null)
                throw new ArgumentNullException("bitmap");

            if (encoder == null)
                throw new ArgumentNullException("encoder");

            encoder.Encode(bitmap, writer);
        }
    }
}
